﻿#include "audio_device_manager.h"

#include "audio/mixer/mixer.h"
#include "audio/mixer/mixer_track.h"
#include "audio/plugin_host/midi_sequencer.h"
#include "audio/plugin_host/plugin_host_manager.h"
#include "spdlog/spdlog.h"
#include "thread_message/thread_message_hubs.h"
#include "misc/platform_thread.h"

IMPL_SINGLETON_INSTANCE(audio_device_manager)

int rt_audio_callback(void* output_buffer, void *input_buffer,
                                unsigned int frames_nums,
                                double stream_time,
                                RtAudioStreamStatus status,
                                void* user_data)
{
    auto* API = static_cast<audio_device_manager*>(user_data);
    return API->stream_callback(static_cast<sample_t*>(output_buffer), static_cast<sample_t*>(input_buffer), frames_nums, stream_time, status);
}

void audio_device_manager::init(singleton_initliazer& initliazer) {
    singleton_t<audio_device_manager>::init(initliazer);
    audio_ = new RtAudio();
    log_all_devices();
    // output_params_.deviceId = audio_->getDefaultOutputDevice();
    output_params_.deviceId = 131;
    output_params_.nChannels = get_output_channel_count();
    output_params_.firstChannel = 0;
    buffer_size_ = 512;
#if USE_DOUBLE_SAMPLE
    spdlog::info("using double sample");
#endif
}

void audio_device_manager::release(singleton_release_guard& release_guard) {
    singleton_t<audio_device_manager>::release(release_guard);
    stop_render_thread();
    release_guard.require_release<mixer>();
    if (audio_)
        audio_->stopStream();
    delete audio_;
}

void audio_device_manager::start() {
    RtAudio::StreamParameters* input_params = nullptr;
    if (input_params_.deviceId != 0) {
        input_params = &input_params_;
    }
    log_current_device_info();

    options_.flags = RTAUDIO_NONINTERLEAVED;
#if USE_DOUBLE_SAMPLE
    RtAudioFormat format = RTAUDIO_FLOAT64;
#else
    RtAudioFormat format = RTAUDIO_FLOAT32;
#endif
    auto err = audio_->openStream(&output_params_, input_params, format, get_sample_rate(), &buffer_size_, &rt_audio_callback, this, &options_);
    if (err != RTAUDIO_NO_ERROR) {
        spdlog::error("failed to open audio stream: {}", audio_->getErrorText());
        return;
    }

    err = audio_->startStream();
    if (err != RTAUDIO_NO_ERROR) {
        spdlog::error("failed to start audio stream: {}", audio_->getErrorText());
        return;
    }
    spdlog::info("audio stream opened");
    start_render_thread();
}

uint32_t audio_device_manager::get_sample_rate() const {
    return sample_rate_;
}

void audio_device_manager::set_sample_rate(uint32_t sample_rate) {
    sample_rate_ = sample_rate;
    on_sample_rate_changed.broadcast(sample_rate_);
}

uint32_t audio_device_manager::get_buffer_size() const {
    return buffer_size_;
}

void audio_device_manager::set_buffer_size(int buffer_size) {
    buffer_size_ = buffer_size;
}

uint32_t audio_device_manager::get_input_channel_count() const {
    return 0;
}

uint32_t audio_device_manager::get_output_channel_count() const {
    return 2; // 现在是固定值, 以后可能会改
}

int audio_device_manager::stream_callback(sample_t* output, sample_t* input, unsigned long frame_count, double stream_time, RtAudioStreamStatus status) {
    if (render_buffers_[0].Num() < frame_count) {
        // spdlog::warn("render buffer underflow: {}", render_buffers_[0].Num());
        return 0;
    }
    for (int i = 0; i < output_params_.nChannels; ++i) {
        render_buffers_[i].Pop(output, frame_count);
        output += frame_count;
    }
    return 0;
}

void audio_device_manager::log_all_devices() {
    const auto& device_count = audio_->getDeviceIds();
    for (const auto& device_id: device_count) {
        const auto& device_info = audio_->getDeviceInfo(device_id);
        spdlog::info("find device id: {}, name: {}", device_id, device_info.name);
    }
}

void audio_device_manager::log_current_device_info() {
    spdlog::info("using output device: {}", audio_->getDeviceInfo(output_params_.deviceId).name);
    spdlog::info("using input device: {}",
                 input_params_.deviceId == 0 ? "none" : audio_->getDeviceInfo(input_params_.deviceId).name);
    spdlog::info("using sample rate: {}", get_sample_rate());
    spdlog::info("using buffer size: {}", get_buffer_size());
    spdlog::info("using output channel count: {}", get_output_channel_count());
    spdlog::info("using input channel count: {}", get_input_channel_count());
}

void audio_device_manager::start_render_thread() {
    render_thread_should_stop_ = false;
    render_thread_ = std::thread(&audio_device_manager::render_thread, this);
    render_thread_.detach();
    spdlog::info("port_audio render thread started");
}

void audio_device_manager::stop_render_thread() {
    render_thread_should_stop_ = true;
    if (render_thread_.joinable())
        render_thread_.join();
    // wait for render thread to stop
    while (render_thread_running_) {
        std::this_thread::sleep_for(std::chrono::milliseconds(1));
    }
    spdlog::info("port_audio render thread stopped");
}

void audio_device_manager::render_thread() {
    set_thread_name("audio_render");
    render_thread_running_ = true;
    dummy_track* master = g_mixer.get_master();

    render_buffers_.resize(get_output_channel_count());

    const uint32_t frames = get_buffer_size();
    const uint32_t rate = get_sample_rate();

    for (auto& render_buffer : render_buffers_)
        render_buffer.SetCapacity(frames * 3);

    const float milliseconds = frames / rate * 1000; // 512 / 48000

    while (!render_thread_should_stop_) {
        g_audio_thread_hub.process_messages();
        if (render_buffers_[0].Remainder() < frames) {
             std::this_thread::sleep_for(std::chrono::milliseconds(static_cast<uint64_t>(milliseconds)));
//            std::this_thread::yield();
            continue;
        }
        g_midi_sequencer.process(get_sample_rate(), frames);
        g_mixer.reset();
        g_plugin_host_manager.process(frames);
        g_mixer.process(frames);

        const auto& master_headers = master->buffer.get_headers_vector();
        for (int i = 0; i < master_headers.size(); ++i) {
            const auto element = master_headers[i];
            render_buffers_[i].Push(element, frames);
        }
    }
    render_thread_running_ = false;
}
