#pragma once

#include <vector>
#include <atomic>

#include "extern.h"
#include "spdlog/spdlog.h"

#if defined(_DEBUG) || defined(DEBUG)
#define checkf(expr, msg, ...) if (!(expr)) { spdlog::error(msg, __VA_ARGS__); throw std::runtime_error(fmt::format(msg, __VA_ARGS__)); }
#define check(expr) if (!(expr)) { spdlog::error("Check failed: {}", #expr); throw std::runtime_error(fmt::format("Check failed: {}", #expr)); }
#else
#define checkf(...)
#define check(...)
#endif

/**
* Basic implementation of a circular buffer built for pushing and popping arbitrary amounts of data at once.
* Designed to be thread safe for SPSC; However, if Push() and Pop() are both trying to access an overlapping area of the buffer,
* One of the calls will be truncated. Thus, it is advised that you use a high enough capacity that the producer and consumer are never in contention.
*/
template<typename SampleType>
class circular_audio_buffer {
private:
    std::vector<SampleType> InternalBuffer;
    uint32_t Capacity;
    std::atomic<int32_t> ReadCounter;
    std::atomic<int32_t> WriteCounter;

public:
    circular_audio_buffer() {
        SetCapacity(0);
    }

    circular_audio_buffer(const circular_audio_buffer& InOther) {
        *this = InOther;
    }

    circular_audio_buffer& operator=(const circular_audio_buffer& InOther) {
        InternalBuffer = InOther.InternalBuffer;
        Capacity = InOther.Capacity;
        ReadCounter.store(InOther.ReadCounter);
        WriteCounter.store(InOther.WriteCounter);

        return *this;
    }


    explicit circular_audio_buffer(uint32_t InCapacity) {
        SetCapacity(InCapacity);
    }

    // move
    circular_audio_buffer(circular_audio_buffer&& InOther) noexcept {
        InternalBuffer = std::move(InOther.InternalBuffer);
        Capacity = InOther.Capacity;
        ReadCounter.store(InOther.ReadCounter);
        WriteCounter.store(InOther.WriteCounter);

        InOther.Capacity = 0;
        InOther.ReadCounter = 0;
        InOther.WriteCounter = 0;
    }

    void Reset(uint32_t InCapacity = 0) {
        SetCapacity(InCapacity);
    }

    void SetCapacity(uint32_t InCapacity) {
        if (InCapacity >= 2147483647) { spdlog::error("Max capacity for this buffer is 2,147,483,647 samples. Otherwise our index arithmetic will not work."); throw std::runtime_error(fmt::format("Max capacity for this buffer is 2,147,483,647 samples. Otherwise our index arithmetic will not work.")); };

        Capacity = InCapacity + 1;
        ReadCounter = 0;
        WriteCounter = 0;
        InternalBuffer.clear();
        InternalBuffer.resize(Capacity);
    }

    /** Reserve capacity.
     *
     * @param InMinimumCapacity - Minimum capacity of circular buffer.
     * @param bRetainExistingSamples - If true, existing samples in the buffer will be retained. If false, they are discarded.
     */
    void Reserve(uint32_t InMinimumCapacity, bool bRetainExistingSamples) {
        if (Capacity <= InMinimumCapacity) {
            uint32_t NewCapacity = InMinimumCapacity + 1;

            checkf(NewCapacity < (uint32_t)INT32_MAX,
                   "Max capacity overflow. Requested {}. Maximum allowed {}", NewCapacity,
                   INT32_MAX);

            uint32_t NumToAdd = NewCapacity - Capacity;
            InternalBuffer.AddZeroed(NumToAdd);
            Capacity = NewCapacity;
        }

        if (!bRetainExistingSamples) {
            ReadCounter = 0;
            WriteCounter = 0;
        }
    }

    /** Push an array of values into circular buffer. */
    int32_t Push(std::vector<const SampleType> InBuffer) {
        return Push(InBuffer.data(), InBuffer.size());
    }

    // Pushes some amount of samples into this circular buffer.
    // Returns the amount of samples written.
    // This can only be used for trivially copyable types.
    int32_t Push(const SampleType* InBuffer, uint32_t NumSamples) {
        SampleType* DestBuffer = InternalBuffer.data();
        const uint32_t ReadIndex = ReadCounter;
        const uint32_t WriteIndex = WriteCounter;

        int32_t NumToCopy = std::min<int32_t>(NumSamples, Remainder());
        const int32_t NumToWrite = std::min<int32_t>(NumToCopy, Capacity - WriteIndex);

        memcpy(&DestBuffer[WriteIndex], InBuffer, NumToWrite * sizeof(SampleType));
        memcpy(&DestBuffer[0], &InBuffer[NumToWrite], (NumToCopy - NumToWrite) * sizeof(SampleType));

        WriteCounter = (WriteIndex + NumToCopy) % Capacity;

        return NumToCopy;
    }

    // Pushes some amount of zeros into the circular buffer.
    // Useful when acting as a blocked, mono/interleaved delay line
    int32_t PushZeros(uint32_t NumSamplesOfZeros) {
        SampleType* DestBuffer = InternalBuffer.data();
        const uint32_t ReadIndex = ReadCounter;
        const uint32_t WriteIndex = WriteCounter;

        int32_t NumToZeroEnd = std::min<int32_t>(NumSamplesOfZeros, Remainder());
        const int32_t NumToZeroBegin = std::min<int32_t>(NumToZeroEnd, Capacity - WriteIndex);

        memset(&DestBuffer[WriteIndex], 0, NumToZeroBegin * sizeof(SampleType));
        memset(&DestBuffer[0], 0, (NumToZeroEnd - NumToZeroBegin) * sizeof(SampleType));

        WriteCounter = (WriteIndex + NumToZeroEnd) % Capacity;

        return NumToZeroEnd;
    }

    // Push a single sample onto this buffer.
    // Returns false if the buffer is full.
    bool Push(const SampleType& InElement) {
        if (Remainder() == 0) {
            return false;
        } else {
            SampleType* DestBuffer = InternalBuffer.data();
            const uint32_t ReadIndex = ReadCounter;
            const uint32_t WriteIndex = WriteCounter;

            DestBuffer[WriteIndex] = InElement;

            WriteCounter = (WriteIndex + 1) % Capacity;
            return true;
        }
    }

    bool Push(SampleType&& InElement) {
        if (Remainder() == 0) {
            return false;
        } else {
            SampleType* DestBuffer = InternalBuffer.data();
            const uint32_t ReadIndex = ReadCounter;
            const uint32_t WriteIndex = WriteCounter;

            DestBuffer[WriteIndex] = MoveTemp(InElement);

            WriteCounter = (WriteIndex + 1) % Capacity;
            return true;
        }
    }

    // Same as Pop(), but does not increment the read counter.
    int32_t Peek(SampleType* OutBuffer, uint32_t NumSamples) const {
        const SampleType* SrcBuffer = InternalBuffer.data();
        const uint32_t ReadIndex = ReadCounter;
        const uint32_t WriteIndex = WriteCounter;

        int32_t NumToCopy = std::min<int32_t>(NumSamples, Num());

        const int32_t NumRead = std::min<int32_t>(NumToCopy, Capacity - ReadIndex);
        memcpy(OutBuffer, &SrcBuffer[ReadIndex], NumRead * sizeof(SampleType));

        memcpy(&OutBuffer[NumRead], &SrcBuffer[0], (NumToCopy - NumRead) * sizeof(SampleType));

        check(NumSamples < ((uint32_t)INT32_MAX));

        return NumToCopy;
    }

    // Peeks a single element.
    // returns false if the element is empty.
    bool Peek(SampleType& OutElement) const {
        if (Num() == 0) {
            return false;
        } else {
            SampleType* SrcBuffer = InternalBuffer.data();
            const uint32_t ReadIndex = ReadCounter;

            OutElement = SrcBuffer[ReadIndex];

            return true;
        }
    }

    // Pops some amount of samples into this circular buffer.
    // Returns the amount of samples read.
    int32_t Pop(SampleType* OutBuffer, uint32_t NumSamples) {
        int32_t NumSamplesRead = Peek(OutBuffer, NumSamples);
        check(NumSamples < ((uint32_t)INT32_MAX));

        ReadCounter = (ReadCounter + NumSamplesRead) % Capacity;

        return NumSamplesRead;
    }

    // Pops some amount of samples into this circular buffer.
    // Returns the amount of samples read.
    int32_t Pop(uint32_t NumSamples) {
        check(NumSamples < ((uint32_t)INT32_MAX));

        int32_t NumSamplesRead = std::min<int32_t>(NumSamples, Num());

        ReadCounter = (ReadCounter + NumSamplesRead) % Capacity;

        return NumSamplesRead;
    }

    // Pops a single element.
    // Will assert if the buffer is empty. Please check Num() > 0 before calling.
    SampleType Pop() {
        // Calling this when the buffer is empty is considered a fatal error.
        check(Num() > 0);

        SampleType* SrcBuffer = InternalBuffer.data();
        const uint32_t ReadIndex = ReadCounter;

        SampleType PoppedValue = std::move(InternalBuffer[ReadIndex]);
        ReadCounter = (ReadCounter + 1) % Capacity;
        return PoppedValue;
    }

    // When called, seeks the read or write cursor to only retain either the NumSamples latest data
    // (if bRetainOldestSamples is false) or the NumSamples oldest data (if bRetainOldestSamples is true)
    // in the buffer. Cannot be used to increase the capacity of this buffer.
    void SetNum(uint32_t NumSamples, bool bRetainOldestSamples = false) {
        check(NumSamples < Capacity);

        if (bRetainOldestSamples) {
            WriteCounter = (ReadCounter + NumSamples) % Capacity;
        } else {
            int64_t ReadCounterNum = ((int32_t) WriteCounter) - ((int32_t) NumSamples);
            if (ReadCounterNum < 0) {
                ReadCounterNum = Capacity + ReadCounterNum;
            }

            ReadCounter = ReadCounterNum;
        }
    }

    // Get number of samples that can be popped off of the buffer.
    uint32_t Num() const {
        const int32_t ReadIndex = ReadCounter;
        const int32_t WriteIndex = WriteCounter;

        if (WriteIndex >= ReadIndex) {
            return WriteIndex - ReadIndex;
        } else {
            return Capacity - ReadIndex + WriteIndex;
        }
    }

    // Get the current capacity of the buffer
    uint32_t GetCapacity() const {
        return Capacity;
    }

    // Get number of samples that can be pushed onto the buffer before it is full.
    uint32_t Remainder() const {
        const uint32_t ReadIndex = ReadCounter;
        const uint32_t WriteIndex = WriteCounter;

        return (Capacity - 1 - WriteIndex + ReadIndex) % Capacity;
    }
};

using circular_buffer_type = circular_audio_buffer<sample_t>;
using circular_buffer_vector_type = std::vector<circular_buffer_type>;

#undef checkf
#undef check