#include "mixer_track.h"

#include "channel_interface.h"
#include "audio/device/audio_device_manager.h"
#include "audio/plugin_host/plugin_host.h"
#include "thread_message/thread_message_hubs.h"

mixer_track::~mixer_track() {
    delete channel_interface_;
    channel_interface_ = nullptr;
    for (auto e : effects) {
        delete_effect(this, e);
    }
}

void mixer_track::init() {
    const uint32_t channel_count = g_audio_device_manager.get_output_channel_count();
    buffer.resize(channel_count, g_audio_device_manager.get_buffer_size());
    for (int i = 0; i < buffer.get_num_channels(); ++i) {
        ui_buffers->emplace_back(g_audio_device_manager.get_buffer_size() * 2);
        ui_buffer_peaks.emplace_back();
    }

    channel_interface_ = new mixer_channel_interface(this);
}

void mixer_track::add_effect(plugin_host* in_effect) {
    g_audio_thread_hub.push_message([in_effect, this] {
        in_effect->owner_tracks.push_back(this);
        in_effect->channel->set_input_channel(channel_interface_->output_channel_nodes);
        in_effect->channel->set_output_channel(channel_interface_->output_channel_nodes);
        effects.push_back(in_effect);
        g_main_thread_hub.push_message([in_effect, this] {
            on_add_effect.broadcast(in_effect);
        });
    });
}

void mixer_track::remove_effect(plugin_host* in_effect) {
    on_remove_effect.broadcast(in_effect);
    g_audio_thread_hub.push_message([in_effect, this] {
        auto remove_effect_track = std::remove(in_effect->owner_tracks.begin(), in_effect->owner_tracks.end(), this);
        in_effect->owner_tracks.erase(remove_effect_track, in_effect->owner_tracks.end());

        auto remove_effect = std::remove(effects.begin(), effects.end(), in_effect);
        effects.erase(remove_effect, effects.end());
    });
}

void mixer_track::add_child(mixer_track* in_child, float in_send_level) {
    children.push_back({in_child, in_send_level});
}

void mixer_track::remove_child(mixer_track* in_child) {
    const auto new_end = std::remove_if(children.begin(), children.end(), [in_child](const mixer_track_link& link) {
        return link.track == in_child;
    });
    children.erase(new_end, children.end());
}

void mixer_track::process(uint32_t in_frames) {
    for (auto effect : effects)
        effect->process(in_frames);
//    buffer.multiple(volume);
    on_processed.broadcast(this);
}

void mixer_track::post_process(uint32_t in_frames) {
    buffer.multiple(get_volume());
    for (int i = 0; i < buffer.get_num_channels(); ++i) {
        auto& ui_buffer = (*ui_buffers)[i];
        ui_buffer.Push(buffer.get_headers()[i], in_frames);
    }
}

void instrument_track::rename(const std::string& in_name) {
    instrument_->name = in_name;
}

std::string instrument_track::get_name() const {
    return instrument_->name;
}

void delete_effect(mixer_track* track, plugin_host* host) {
    track->remove_effect(host);
    host->try_close_editor();
    g_audio_thread_hub.push_message([host](){
        delete host;
    });
}
